#!/usr/bin/python
# Copyright 2008 Canonical Ltd.  All rights reserved.

"""Query our buildbot and do stuff.

    - Switch the current pqm config depending on if the last buildbot run
      failed or not.
    - Push the last known good revision of our development branch to our
      stable branch.
    - Optionally send a request to pqm to merge stable into the db-devel
      branch.
"""

from optparse import OptionParser
import os.path
import subprocess
import sys
from urllib2 import (
    build_opener, HTTPCookieProcessor, HTTPError, HTTPRedirectHandler,
    Request)
from urlparse import urljoin, urlparse, urlunparse
import warnings
import xmlrpclib

import bzrlib.errors
import bzrlib.bzrdir
import bzrlib.missing
from bzrlib import gpg
from bzrlib import config
from bzrlib.plugins.pqm import pqm_submit
from bzrlib.smtp_connection import SMTPConnection


class GPGStrategy(gpg.GPGStrategy):

    _config = None

    def __init__(self, user=None, passphrase_file=None):
        self.user = user
        self.passphrase_file = passphrase_file

    def _command_line(self):
        res =  ['gpg']
        if self.user is not None:
            res.extend(['-u', self.user])
        if self.passphrase_file is not None:
            res.extend(['--passphrase-file', self.passphrase_file])
        res.append('--clearsign')
        res.append('--no-tty')
        return res


class PQMSubmission(pqm_submit.PQMSubmission):

    tree = source_branch=None

    def __init__(self, public_location, message, submit_location,
                 user=None, passphrase_file=None):
        self.public_location = public_location
        self.submit_location = submit_location
        self.user = user
        self.passphrase_file = passphrase_file
        # Check that the message is okay to pass to PQM
        assert message is not None
        self.message = message.encode('utf8')
        if '\n' in self.message:
            raise pqm_submit.BadCommitMessage(self.message)

    def to_signed(self):
        """Serialize as a signed string."""
        unsigned_text = ''.join(self.to_lines())
        unsigned_text = unsigned_text.encode('ascii') #URLs should be ascii

        strategy = GPGStrategy(user=self.user,
                               passphrase_file=self.passphrase_file)
        return strategy.sign(unsigned_text)


# We need tempnam to do an atomic symlink.
warnings.filterwarnings(
    "ignore", "tempnam is a potential security risk",
    RuntimeWarning)


def homepath(*path_components):
    """Construct a path relative to $HOME."""
    return os.path.join(os.environ['HOME'], *path_components)


# Command line argument defaults.
DEFAULTS = {
    'buildbot': 'http://lpbuildbot.canonical.com:8010',
    'builder': 'lucid_lp',
    'dest_config': homepath('pqm-config', 'launchpad.conf'),
    'success_config': homepath('pqm-config', 'launchpad-success.conf'),
    'failure_config': homepath('pqm-config', 'launchpad-failure.conf'),
    'devel_branch': homepath('archives', 'rocketfuel', 'launchpad', 'devel'),
    'stable_branch': 'lp:~launchpad-pqm/launchpad/stable',
    'quiet': False,
    # these bits are for submitting stable to db-devel
    'db_devel_branch': 'lp:~launchpad-pqm/launchpad/db-devel',
    'gpg_user': None,
    'gpg_passphrase_file': None,
    'email': 'Launchpad PQM <launchpad@lists.canonical.com>',
    'pqm_email': 'Launchpad PQM <launchpad@pqm.canonical.com>',
    'dry_run': False,
    # these bits are for checking on the db branches
    'db_builder': 'lucid_db_lp',
    'db_devel_local_branch': homepath(
        'archives', 'rocketfuel', 'launchpad', 'db-devel'),
    'db_stable_branch': 'lp:~launchpad-pqm/launchpad/db-stable',
    }


# XMLRPCRedirectHandler and UrlLib2Transport taken from Launchpad's
# canonical.launchpad.components.externalbugtracker.xmlrpc module.
# It is not imported from that location because this script needs
# to be standalone.
class XMLRPCRedirectHandler(HTTPRedirectHandler):
    """A handler for HTTP redirections of XML-RPC requests."""

    def redirect_request(self, req, fp, code, msg, headers, newurl):
        """Return a Request or None in response to a redirect.

        See `urllib2.HTTPRedirectHandler`.

        If the original request is a POST request, the request's payload
        will be preserved in the redirect and the returned request will
        also be a POST request.
        """
        # If we can't handle this redirect,
        # HTTPRedirectHandler.redirect_request() will raise an
        # HTTPError. We call the superclass here in the old fashion
        # since HTTPRedirectHandler isn't a new-style class.
        new_request = HTTPRedirectHandler.redirect_request(
            self, req, fp, code, msg, headers, newurl)

        # If the old request is a POST request, the payload will be
        # preserved. Note that we don't need to test for the POST-ness
        # of the old request; if its data attribute - its payload - is
        # not None it's a POST request, if it's None it's a GET request.
        # We can therefore just copy the data from the old request to
        # the new without worrying about breaking things.
        new_request.data = req.data
        return new_request


class UrlLib2Transport(xmlrpclib.Transport):
    """An XMLRPC transport which uses urllib2.

    This XMLRPC transport uses the Python urllib2 module to make the
    request, and connects via the HTTP proxy specified in the
    environment variable `http_proxy`, i present. It also handles
    cookies correctly, and in addition allows specifying the cookie
    explicitly by setting `self.auth_cookie`.

    Note: this transport isn't fit for general XML-RPC use. It is just
    good enough for some of our extrnal bug tracker implementations.

    :param endpoint: The URL of the XMLRPC server.
    """

    verbose = False

    def __init__(self, endpoint, cookie_jar=None):
        # Expected by Python 2.5+.
        self._use_datetime = 0
        self.scheme, self.host = urlparse(endpoint)[:2]
        assert self.scheme in ('http', 'https'), (
            "Unsupported URL schene: %s" % self.scheme)
        self.cookie_processor = HTTPCookieProcessor(cookie_jar)
        self.redirect_handler = XMLRPCRedirectHandler()
        self.opener = build_opener(
            self.cookie_processor, self.redirect_handler)

    def setCookie(self, cookie_str):
        """Set a cookie for the transport to use in future connections."""
        name, value = cookie_str.split('=')
        cookie = Cookie(
            version=0, name=name, value=value,
            port=None, port_specified=False,
            domain=self.host, domain_specified=True,
            domain_initial_dot=None,
            path='', path_specified=False,
            secure=False, expires=False, discard=None,
            comment=None, comment_url=None, rest=None)
        self.cookie_processor.cookiejar.set_cookie(cookie)

    def request(self, host, handler, request_body, verbose=0):
        """Make an XMLRPC request.

        Uses the configured proxy server to make the connection.
        """
        url = urlunparse((self.scheme, host, handler, '', '', ''))
        headers = {'Content-type': 'text/xml'}
        request = Request(url, request_body, headers)
        try:
            response = self._parse_response(self.opener.open(request), None)
        except HTTPError, he:
            raise xmlrpclib.ProtocolError(
                request.get_full_url(), he.code, he.msg, he.hdrs)
        return response


def main():
    parser = OptionParser()
    parser.add_option(
        '--buildbot', dest='buildbot',
        help="Buildbot can be found at URL", metavar="URL")
    parser.add_option(
        '--builder', dest='builder',
        help="Builder is called NAME", metavar="NAME")
    parser.add_option(
        '--dest-config', dest="dest_config", metavar="PATH",
        help="PATH will be replaced with a symlink to either the success "
             "or failure config")
    parser.add_option(
        '--success-config', dest="success_config", metavar="PATH",
        help="The config to use when buildbot is succeeding is at PATH")
    parser.add_option(
        '--failure-config', dest="failure_config", metavar="PATH",
        help="The config to use when buildbot is failing is at PATH")
    parser.add_option(
        '--devel-branch', dest='devel_branch', metavar="PATH",
        help="The branch being built by buildbot is on the local filesystem "
             "at PATH")
    parser.add_option(
        '--stable-branch', dest="stable_branch", metavar="BRANCH",
        help="Last known good build is pushed to BRANCH. "
             "if this is a local directory it should be an absolute path.")
    parser.add_option(
        '--force-result', dest="force_result", metavar="RESULT",
        help="Force result RESULT on the last build, ignoring the "
             "real result.")
    parser.add_option(
        '-q', '--quiet', dest="quiet", action="store_true",
        help="No informative output on success.")
    # these bits are for submitting stable to db-devel
    parser.add_option(
        '--db-devel-branch', dest='db_devel_branch', metavar='DBBRANCH',
        help='Generates a PQM merge request to merge stable branch to this '
             'branch.')
    parser.add_option(
        '--gpg-user', dest='gpg_user', metavar='GPGUSER',
        help='The GPG user for PQM submission (using standard gpg syntax).')
    parser.add_option(
        '--gpg-passphrase-file', dest='gpg_passphrase_file',
        metavar='PASSPHRASEFILE',
        help='The file containing the gpg passphrase.')
    parser.add_option(
        '--pqm-email', dest='pqm_email', metavar='PQMEMAIL',
        help='The email address of PQM.')
    parser.add_option(
        '--email', dest='email', metavar='USEREMAIL',
        help='The email address from which the PQM email will claim to have '
             'been sent.')
    # these bits are for checking on the db branches
    parser.add_option(
        '--db-builder', dest='db_builder',
        help="DB Builder is called DBNAME", metavar="DBNAME")
    parser.add_option(
        '--db-devel-local-branch', dest='db_devel_local_branch',
        metavar="DBPATH",
        help="The db branch being built by buildbot is on the local "
             "filesystem at DBPATH")
    parser.add_option(
        '--db-stable-branch', dest="db_stable_branch", metavar="DBBRANCH",
        help="Last known good db build is pushed to DBBRANCH. "
             "if this is a local directory it should be an absolute path.")
    # debug option
    parser.add_option(
        '--dry-run', dest='dry_run', action='store_true',
        metavar='DRYRUN',
        help='Do not do anything, just report output.')
    parser.set_defaults(**DEFAULTS)

    (options, args) = parser.parse_args()

    if len(args) > 0:
        parser.error("Too many arguments.")

    if (os.path.exists(options.dest_config)
        and not os.path.islink(options.dest_config)):
        parser.error(
            "--dest-config %s is not a symbolic link. Not overwriting."
            % options.dest_config)

    if not os.path.isfile(options.success_config):
        parser.error(
            "--success-config %s does not exist." % options.success_config)

    if not os.path.isfile(options.failure_config):
        parser.error(
            "--failure-config %s does not exist." % options.failure_config)

    if not os.path.isdir(options.devel_branch):
        parser.error(
            "--devel-branch %s is not a local directory."
            % options.devel_branch)

    if not os.path.isdir(options.db_devel_local_branch):
        parser.error(
            "--db-devel-local-branch %s is not a local directory."
            % options.db_devel_local_branch)

    buildbot_endpoint = urljoin(options.buildbot, 'xmlrpc')
    buildbot = xmlrpclib.ServerProxy(
        buildbot_endpoint, UrlLib2Transport(buildbot_endpoint))

    revision, result = check_builder_and_push_to_stable(
        buildbot, options.builder, options.devel_branch,
        options.stable_branch, options)

    db_revision, db_result = check_builder_and_push_to_stable(
        buildbot, options.db_builder, options.db_devel_local_branch,
        options.db_stable_branch, options)

    current_build_reasons = buildbot.getReasonsForRunningBuilds()

    # If both builds are successful, or they have received a [testfix]
    # landing, we should be in normal mode.
    if (is_revision_clean(
        result, revision, options.devel_branch, options,
        current_build_reasons['lucid_lp'])
        and is_revision_clean(
                db_result, db_revision, options.db_devel_local_branch,
                options, current_build_reasons['lucid_db_lp'])):
        set_pqm_mode('normal', options)
    else:
        # Otherwise, we are in testfix mode.
        set_pqm_mode('testfix', options)

    # If both builds succeeded, request a merge
    if db_result == result == 'success' and options.db_devel_branch:
        make_merge_request(options, revision)

    return 0


def check_builder_and_push_to_stable(buildbot, builder,
                                     local_devel_branch,
                                     stable_branch, options):
    """Check a builder and push to the stable branch in case of success.

    Return the build revision and the status result.
    """
    # Get the last 10 builds info. We request more build info for the case
    # where the last one actually failed on checkout. We need a couple of
    # backlog to see what is the last revision actually tested.
    latest_build_infos = buildbot.getLastBuilds(builder, 10)

    # They are sorted older first.
    (builder, build_number, build_start, build_end,
        branch_name, revision, result, text, reasons) = latest_build_infos.pop()

    if options.force_result is not None:
        result = options.force_result

    try:
        revision = int(revision)
    except ValueError:
        if revision in ('None', ''):
            # The last build failed before checkout. Find the latest revision
            # tested before that.
            if result not in ('failure', 'exception'):
                raise AssertionError("No revision but not a failure!")
            if not options.quiet:
                print ('Builder %s at %s failed to checkout branch '
                    'during last build: %s' % (
                    builder, options.buildbot, " ".join(text)))
            for info in reversed(latest_build_infos):
                try:
                    return int(info[4]), result
                except ValueError:
                    # Ignore invalid revision.
                    pass
            else:
                print ("ERROR: The last 10 builds failed on checkout. "
                    "Something is seriously broken!")
                sys.exit(1)
        else:
            if not options.quiet:
                print ("ERROR: Latest build on %s doesn't contain an integer "
                      "revision [%r]" % (builder, revision,))
            sys.exit(1)

    # Push to stable on success.
    if result == 'success':
        if not options.quiet:
            print 'Builder %s at %s succeeded building r%s.' % (
                builder, options.buildbot, revision)
        on_build_succeeded(
            local_devel_branch, stable_branch, revision, options.quiet,
            options.dry_run)
    else:
        if not options.quiet:
            print 'Builder %s at %s failed building r%s [%s].' % (
                builder, options.buildbot, revision, " ".join(text))

    return revision, result


def is_revision_clean(result, revision, local_branch, options,
                      current_build_reasons):
    """The revision is clean for testfix mode in three cases:

        * the build was successful
        * a build was forced in the buildbot UI
        * or there was a subsequent revision with [testfix] in it.

    We assume that builds are only forced in situations where the failure was
    transient (caused by some network glitch, for example) and that any
    [testfix] will be sufficient in getting us of out of testfix mode.

    Note that the branch isn't pushed to stable until there is a successful
    build.
    """
    if result == 'success':
        return True

    # If a build was forced in the UI, assume the failure was spurious and the
    # branch is clean.
    if 'build forced' in current_build_reasons:
        return True

    # Otherwise, check that there is a revision with '[testfix]' in its
    # commit message.
    branch = bzrlib.branch.Branch.open(local_branch)
    failed_rev_id = branch.get_rev_id(revision)
    branch.lock_read()
    try:
        for revid, depth, revno, end in branch.iter_merge_sorted_revisions(
            stop_revision_id=failed_rev_id, direction="forward"):
            # Only look at mainline revisions.
            if depth != 0:
                continue
            message = branch.repository.get_revision(revid).message
            if message.strip().startswith('[testfix]'):
                if not options.quiet:
                    print "Found [testfix] on revision %s." % revno[0]
                return True
    finally:
        branch.unlock()
    return False


def normalize_branch_name(branch_name):
    """Normalize lp: branch name into something that PQM groks."""
    if branch_name.startswith('lp:'):
        return 'bzr+ssh://bazaar.launchpad.net/' + branch_name[3:]
    else:
        return branch_name


def make_merge_request(options, revision):
    """Send a merge request from stable to db_devel through PQM."""

    # Unlike pushes (as above), a pqm submission can take a noticeably
    # long time--ten minutes or so--and if it is unnecessary, it can
    # slow down developers. So we only want to generate a submission if
    # the db_devel branch needs our current revision. We could stash the
    # last-submitted revision around locally somewhere, but that's
    # fragile. It's much nicer if we can actually ask db_devel if it has
    # the revision we we should try to generate a pqm submission.  So
    # that's what we'll do with bzrlib.
    stable_branch = normalize_branch_name(options.stable_branch)
    db_devel_branch = normalize_branch_name(options.db_devel_branch)
    bzr_open = bzrlib.bzrdir.BzrDir.open_containing_tree_or_branch
    if not options.quiet:
        print 'opening ' + stable_branch
    tree, branch, path = bzr_open(stable_branch)
    if not options.quiet:
        print 'opening ' + db_devel_branch
    db_tree, db_branch, db_path = bzr_open(db_devel_branch)
    if not options.quiet:
        print 'checking if db-devel branch is up-to-date with stable'
    # Spelling thanks to abentley:
    # branch2.repository.get_graph(branch1.repository).is_ancestor(
    #     revision_id, branch2.last_revision())
    revid = branch.get_rev_id(revision)
    graph = db_branch.repository.get_graph(branch.repository)
    if not graph.is_ancestor(revid, db_branch.last_revision()):
        if not options.quiet:
            print 'preparing PQM merge request.'
        # Figure out which revisions we're landing from stable.
        missing_revs = bzrlib.missing.find_unmerged(
            branch, db_branch, restrict='local')[0]
        stable_revnos = [revno for revno, rev in missing_revs]
        submission = PQMSubmission(
            stable_branch,
            ('[rs=buildbot-poller] automatic merge from stable. '
            'Revisions: %s included.' % ', '.join(stable_revnos)),
            db_devel_branch, options.gpg_user or options.email,
            options.gpg_passphrase_file
            )

        message = submission.to_email(
            options.email.encode('utf8'),
            options.pqm_email.encode('utf8'))

        if options.dry_run:
            print message.as_string()
        else:
            SMTPConnection(config.GlobalConfig()).send_email(message)
    else:
        # the desired revision has already been merged.
        if not options.quiet:
            print '(no PQM request needed)'


def get_pqm_mode(options):
    """Return the current PQM mode based on the current config it is using."""
    if os.path.samefile(options.success_config, options.dest_config):
        return "normal"
    elif os.path.samefile(options.failure_config, options.dest_config):
        return 'testfix'
    else:
        return 'unknown'


def set_pqm_mode(mode, options):
    """Change the PQM configuration based on the mode we are in."""
    current_mode = get_pqm_mode(options)
    if current_mode == mode:
        return

    if mode == 'normal':
        source_path = options.success_config
    elif mode == 'testfix':
        source_path = options.failure_config
    else:
        raise AssertionError("Unknown mode: %s" % mode)
    if not options.quiet:
        print "Switching to %s mode" % mode

    # Use tempnam to avoid the race condition in doing unlink/symlink.
    temporary_name = os.tempnam(os.path.dirname(options.dest_config))
    if options.dry_run:
        print 'DRY RUN: SYMLINK %s TO %s' % (source_path, temporary_name)
        print 'DRY RUN: RENAME %s TO %s' % (
            temporary_name, options.dest_config)
    else:
        os.symlink(source_path, temporary_name)
        os.rename(temporary_name, options.dest_config)


def on_build_succeeded(local_devel_branch, stable_branch, revision, quiet,
                       dry_run):
    cmd = ['bzr', 'push', '-r', str(revision), stable_branch]
    if dry_run:
        bzr_stdout = 'DRY RUN: FROM %s DO CMD %s' % (local_devel_branch,
                                                     ' '.join(cmd))
        failure = False
    else:
        bzr_process = subprocess.Popen(
            cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT, shell=False, cwd=local_devel_branch)
        (bzr_stdout, bzr_stderr) = bzr_process.communicate()
        failure = bzr_process.returncode != 0
    if failure:
        print 'ERROR: %s failed [%d]' % (
            ' '.join(cmd), bzr_process.returncode)
        print bzr_stdout
    else:
        if not quiet:
            print bzr_stdout


if __name__ == '__main__':
    sys.exit(main())
