import bzrlib.branch
import twisted.python.log
import twisted.internet.defer
import twisted.internet.reactor
import twisted.internet.task
import twisted.internet.threads

import buildbot.util
import buildbot.changes.base
import buildbot.changes.changes

import bzrbuildbot.change


class BzrPoller(buildbot.changes.base.ChangeSource,
                buildbot.util.ComparableMixin):

    compare_attrs = ['urls']

    def __init__(self, urls, poll_interval=10*60, blame_merge_author=False):
        # poll_interval is in seconds, so default poll_interval is 10 minutes
        # bzr+ssh://bazaar.launchpad.net/~launchpad-pqm/launchpad/devel/
        # works, lp:~launchpad-pqm/launchpad/devel/ doesn't.
        if isinstance(urls, basestring):
            urls = (urls,)
        self.last_revisions = {}
        for url in urls:
            if url.startswith('lp:'):
                url = 'bzr+ssh://bazaar.launchpad.net/' + url[3:]
            self.last_revisions[url] = None
        self.poll_interval = poll_interval
        self.loop = twisted.internet.task.LoopingCall(self.poll)
        self.blame_merge_author = blame_merge_author

    @property
    def urls(self):
        return self.last_revisions.keys()

    def startService(self):
        twisted.python.log.msg("BzrPoller%r starting" % self.urls)
        buildbot.changes.base.ChangeSource.startService(self)
        twisted.internet.reactor.callWhenRunning(
            self.loop.start, self.poll_interval)
        missing = len([v for v in self.last_revisions.values() if v is None])
        for change in reversed(self.parent.changes):
            if not missing:
                break
            if self.last_revisions.get(change.branch, self) is None:
                self.last_revisions[change.branch] = change.revision
                missing -= 1
        self.polling = False

    def stopService(self):
        twisted.python.log.msg("BzrPoller%r shutting down" % self.urls)
        self.loop.stop()
        return buildbot.changes.base.ChangeSource.stopService(self)

    def describe(self):
        return "BzrPoller watching %r" % self.urls

    @twisted.internet.defer.inlineCallbacks
    def poll(self):
        if self.polling: # this is called in a loop, and the loop might
            # conceivably overlap.
            return
        self.polling = True
        try:
            # this was concurrent, but bzr doesn't seem to like concurrent
            # connections very much.  So we serialize.
            for url in self.urls:
                yield self._poll(url)
        finally:
            self.polling = False

    @twisted.internet.defer.inlineCallbacks
    def _poll(self, url):
        # On a big tree, even individual elements of the bzr commands can
        # take awhile.  So we just push the bzr work off to a thread.
        try:
            changes = yield twisted.internet.threads.deferToThread(
                self.getRawChanges, url)
        except (SystemExit, KeyboardInterrupt):
            raise
        except:
            # we'll try again next poll.  Meanwhile, let's report.
            twisted.python.log.err()
        else:
            for change in changes:
                change['branch'] = url
                yield self.addChange(
                    buildbot.changes.changes.Change(**change))
                self.last_revisions[url] = change['revision']

    def getRawChanges(self, url):
        branch = bzrlib.branch.Branch.open_containing(url)[0]
        try:
            changes = []
            change = bzrbuildbot.change.generate(
                branch, blame_merge_author=self.blame_merge_author)
            last_revision = self.last_revisions[url]
            if (last_revision is None or
                change['revision'] > last_revision):
                changes.append(change)
                if last_revision is not None:
                    while last_revision + 1 < change['revision']:
                        change = bzrbuildbot.change.generate(
                            branch, new_revno=change['revision']-1,
                            blame_merge_author=self.blame_merge_author)
                        changes.append(change)
            changes.reverse()
            return changes
        finally:
            # Ensure that all transports are disconnected.
            branch.bzrdir.root_transport.disconnect()

    def addChange(self, change):
        d = twisted.internet.defer.Deferred()
        def _add_change():
            d.callback(
                self.parent.addChange(change))
        twisted.internet.reactor.callLater(0, _add_change)
        return d
