# Copyright 2008 Canonical Ltd.  All rights reserved.

import datetime
import os
import time

import twisted.application.service
import twisted.python.log
import twisted.internet.reactor

import buildbot.buildset
import buildbot.changes.changes
import buildbot.scheduler
import buildbot.sourcestamp
import buildbot.status.builder
import buildbot.util


class AggregatingTestfix(buildbot.scheduler.BaseUpstreamScheduler):
    # states:
    # - failure ("[testfix]"):
    #   * we have a change with "[testfix]" in the message: run build
    #     immediately.
    #   * we do not have a "[testfix]" change: do nothing.
    # - non-failure:
    #   * treeStableTimer is None; or treeStableCount is set and change count
    #     >= treeStableCount; or treeStableTimer is set, and time since last
    #     change >= treeStableTimer; or timer has fired: run tests immediately
    #   * treeStableTimer is not None; and treeStableCount is None or change
    #     count < treeStableCount; and time since last change <
    #     treeStableTimer: set Timer

    timer = nextBuildTime = None

    def __init__(self, name, builderNames, branch, treeStableTimer=None,
                 treeStableCount=None, properties={}):
        buildbot.scheduler.BaseUpstreamScheduler.__init__(
            self, name, properties)
        self.branch = branch
        self.changes = []
        self.builderNames = builderNames
        self._submission_time = None
        self.failing = False
        self.treeStableTimer = treeStableTimer
        if treeStableTimer is None and treeStableCount is not None:
            raise ValueError(
                'You may only set treeStableCount if treeStableTimer is not '
                'None')
        self.treeStableCount = treeStableCount

    def listBuilderNames(self):
        return self.builderNames

    def getPendingBuildTimes(self):
        if self.nextBuildTime is not None:
            return [self.nextBuildTime]
        return []

    def _isInterestingChange(self, change):
        return (
            self.branch is None or (
                change.branch is not None and
                change.branch.endswith(self.branch)))

    def addChange(self, change):
        if not self._isInterestingChange(change):
            return
        self.changes.append(change)
        if self._submission_time is None:
            self._maybeSubmitChanges()

    def _maybeSubmitChanges(self):
        self.stopTimer()
        if self.failing:
            for change in self.changes:
                if change.comments and '[testfix]' in change.comments.lower():
                    self._submitChanges()
            else:
                # in failing mode, and no testfix.  Will not schedule.
                twisted.python.log.msg(
                    '%s in failing mode and no "[testfix]" in comment, '
                    'so no build' % (self,))
            self.nextBuildTime = None
            return
        submit = False
        if self.treeStableTimer is not None:
            change = self.changes[-1]
            self.nextBuildTime = max(self.nextBuildTime,
                                     change.when + self.treeStableTimer)
            if (self.treeStableCount and
                len(self.changes) >= self.treeStableCount):
                    submit = True
        else:
            submit = True
        if submit:
            self.nextBuildTime = None
            self._submitChanges()
        else:
            # If nextBuildTime is in the past, will schedule for almost
            # immediately.
            self.setTimer(self.nextBuildTime)

    def _submitChanges(self, next_changes=None):
        # create a BuildSet, submit it to the BuildMaster
        bs = buildbot.buildset.BuildSet(
            self.builderNames,
            buildbot.sourcestamp.SourceStamp(changes=self.changes),
            properties=self.properties)
        self._submission_time = datetime.datetime.utcnow()
        self.changes = next_changes or []
        self.submitBuildSet(bs)

    def buildSetFinished(self, bss):
        buildbot.scheduler.BaseUpstreamScheduler.buildSetFinished(self, bss)
        self._submission_time = None
        self.failing = bss.getResults() != buildbot.status.builder.SUCCESS
        if self.failing:
            twisted.python.log.msg(
                '%s last build did not succeed (%r)' % (self, bss.getResults()))
        if self.changes:
            self._maybeSubmitChanges()

    def setTimer(self, when):
        twisted.python.log.msg("%s: setting timer to %s" %
                (self, time.strftime("%H:%M:%S", time.localtime(when))))
        now = buildbot.util.now()
        if when < now:
            when = now + 1
        if self.timer:
            self.timer.cancel()
        self.timer = twisted.internet.reactor.callLater(
            when - now, self.fireTimer)

    def stopTimer(self):
        if self.timer:
            self.timer.cancel()
            self.timer = None

    def fireTimer(self):
        # clear out our state
        self.timer = None
        self.nextBuildTime = None
        self._submitChanges()

    def startService(self):
        buildbot.scheduler.BaseUpstreamScheduler.startService(self)
        # this assumes that this scheduler's builders do not get jobs from
        # other schedulers.
        #
        # supply a BUILDBOT_IGNORE_LAST_FAILURE environment variable (any
        # value) to a fresh start (or restart) of the buildbot if you want
        # buildbot to ignore a previous test failure.  This allows you to
        # force-ignore a previous test failure.
        ignore_last_failure = bool(
            os.environ.get('BUILDBOT_IGNORE_LAST_FAILURE'))
        rerun_last_failure = bool(
            os.environ.get('BUILDBOT_RERUN_LAST_FAILURE'))
        last_revision = last_changes = None
        for name in self.builderNames:
            builder = self.parent.status.getBuilder(name)
            last_build = builder.getLastFinishedBuild()
            if last_build is not None and last_build.changes:
                # get the last_revision of the last run.  Slaves kill active
                # runs when their master dies, so this should be all we care
                # about, without having to look at active and pending builds.
                # build.changes[-1] is most recent change in a build.
                revision = last_build.changes[-1].revision
                if last_revision is None or revision < last_revision:
                    last_revision = revision
            # maybe check whether last build failed
            if (not ignore_last_failure and
                not self.failing and last_build is not None and
                last_build.getResults() != buildbot.status.builder.SUCCESS):
                # the previous build failed, and we want to remember that
                # and be in [testfix] mode
                self.failing = True
                twisted.python.log.msg(
                    '%s remembering previous failure' % (self,))
                if rerun_last_failure:
                    last_changes = last_build.changes[:]
        if last_revision is not None:
            # get all the interesting changes since the last revision
            changes = []
            for change in reversed(self.parent.change_svc.changes):
                if (change.revision > last_revision and
                    self._isInterestingChange(change)):
                    # this change should be part of a new build request
                    changes.append(change)
                else:
                    break
            changes.reverse()
            if rerun_last_failure and last_changes:
                twisted.python.log.msg(
                    '%s rescheduling last failed build' % (self,))
                self.changes = last_changes
                twisted.internet.reactor.callWhenRunning(
                    self._submitChanges, changes)
            elif changes:
                # if there were any changes, maybe start a build once the
                # reactor has started
                self.changes = changes
                twisted.internet.reactor.callWhenRunning(
                    self._maybeSubmitChanges)
            if changes:
                twisted.python.log.msg(
                    '%s remembering %d untested changes' %
                    (self, len(changes)))

    def stopService(self):
        self.stopTimer()
        return twisted.application.service.MultiService.stopService(self)
