import buildbot.buildset
import buildbot.sourcestamp
from buildbot.status import html
from buildbot.status.web import xmlrpc

from bzrbuildbot.scheduler import AggregatingTestfix

from twisted.web.resource import Resource
from twisted.web.error import ErrorPage


force_html = '''
<html>
 <head>
  <title>Force a build</title>
 </head>
 <body>
  <h1>Force a build</h1>
  <form action="force" method="post">
   <ul>
    %(buttons)s
   </ul>
   <input name="force" type="submit" value="force">
  </form>
 </body>
</html>
'''

button_html = '''
<li>
 <input type="radio" name="build" value="%(index)s"> Build %(branch)s on %(builder)s
</li>
'''

success_html = '''
<html>
 <head>
  <title>Build forced!</title>
 </head>
 <body>
  <a href=".">Carry on</a>
 </body>
</html>
'''

def bad_request_with_message(request, detail="Forcing the build didn't work"):
    real_detail = '<p>%s</p><p><a href="force">Try again</a></p>' % detail
    return ErrorPage(400, "Something went wrong", real_detail).render(request)

class ForceResource(Resource):

    def __init__(self, forceableBranchBuilderNames):
        self.forceableBranchBuilderNames = forceableBranchBuilderNames

    def render_GET(self, request):
        buttons = []
        for index, (branch, builder) in enumerate(self.forceableBranchBuilderNames):
            buttons.append(
                button_html %
                {'builder': builder, 'index':index, 'branch': branch})
        return force_html % {'buttons': '\n'.join(buttons)}

    def render_POST(self, request):
        builds = request.args.get('build')
        if builds:
            try:
                build = int(builds[0])
            except ValueError:
                return bad_request_with_message(
                    request, "non integer build %r"%build)
            try:
                branch, builder = self.forceableBranchBuilderNames[build]
            except IndexError:
                return bad_request_with_message(
                    request, "out of bounds build %r"%build)
            self._submitBuild(branch, builder)
            return success_html
        return bad_request_with_message(
                    request, "no build selected")

    def _submitBuild(self, branch, builder):
        reason = 'build forced'
        relevant_schedulers = [
            scheduler for scheduler in self.master.allSchedulers()
            if builder in scheduler.listBuilderNames()
            and isinstance(scheduler, AggregatingTestfix)]
        if relevant_schedulers:
            for scheduler in relevant_schedulers:
                bs = buildbot.buildset.BuildSet(
                    [builder], buildbot.sourcestamp.SourceStamp(
                        branch=branch, changes=scheduler.changes),
                    reason=reason)
                scheduler.changes = []
                scheduler.submitBuildSet(bs)
        else:
            bs = buildbot.buildset.BuildSet(
                [builder], buildbot.sourcestamp.SourceStamp(branch=branch),
                reason=reason)
            self.master.submitBuildSet(bs)


class XMLRPCServer(xmlrpc.XMLRPCServer):
    def xmlrpc_getReasonsForRunningBuilds(self):
        r = {}
        for builder_name in self.status.getBuilderNames():
            r[builder_name] = reasons = []
            builder = self.status.getBuilder(builder_name)
            for build in builder.currentBuilds:
                reasons.append(build.reason)
        return r


class WebStatus(html.WebStatus):

    def __init__(self, *args, **kw):
        forceableBranchBuilderNames = kw.pop(
            'forceableBranchBuilderNames', [])
        html.WebStatus.__init__(self, *args, **kw)
        self.force_resource = ForceResource(forceableBranchBuilderNames)
        self.putChild('force', self.force_resource)
        self.putChild('xmlrpc', XMLRPCServer())

    def setServiceParent(self, parent):
        html.WebStatus.setServiceParent(self, parent)
        self.force_resource.master = parent
